# Generated by Django 2.2.11 on 2020-03-23 15:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0005_optimize_sync'),
    ]

    operations = [
        migrations.AddField(
            model_name='updatecollectionpackage',
            name='relogin_suggested',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='updatecollectionpackage',
            name='restart_suggested',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='updatecollection',
            name='name',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='updatecollection',
            name='shortname',
            field=models.TextField(null=True),
        ),
    ]
